from pathlib import Path
from collections import namedtuple

import numpy
from PIL import Image, ImageDraw, ImageFont

Point = namedtuple("Point", ["x", "y"])
Spot = namedtuple("Spot", ["found", "x", "y"])


THIS_DIR = Path(__file__).absolute().parent
FONT_PATH = THIS_DIR / "arial.ttf"
FONT = ImageFont.truetype(str(FONT_PATH), 32)

GREEN = (0, 255, 0)
RED = (255, 0, 0)


def recalculate(iterable, factor):
    """dirty hack to apply a factor to items in a tuple

    iterable: the tuple to apply the factor
    factor: the factor to apply

    returns: tuple or namedtuple with factor applied to items
    """
    recalculated = [v * factor for v in iterable]
    try:
        # for named tuples
        cls = type(iterable)
        return cls(*recalculated)
    except TypeError:
        # for tuples
        return tuple(recalculated)


def convert_16bit_to_8bit(img):
    """convert a 16 bit grey scale image to 8 bit without intensity crop

    img: image, 16 bit greyscale

    returns: image, 8 bit greyscale
    """
    array = numpy.uint8(numpy.array(img) / 256)
    return Image.fromarray(array)


def convert_16bit_grey_to_color(img):
    """convert a 16 bit grey scale image to 8 bit rgbwithout intensity crop

    img: image, 16 bit greyscale

    returns: image, 8 bit rgb
    """
    return convert_16bit_to_8bit(img).convert(mode="RGB")


def resize(img, scale=1):
    """resizes an image by a scale factor

    img: image to resize
    scale: scale factor

    returns: rescaled image
    """
    return img.resize(recalculate(img.size, scale))


def load_array_image(file_path, scale=1):
    """loads an array image file

    The image is converted to rgb and scaled.

    file_path: path to image file
    scale: factor for scaling the image

    returns: image
    """
    img = Image.open(file_path)
    colored = convert_16bit_grey_to_color(img)
    return resize(colored, scale)


def get_position(spot_data, actual=True):
    """returns the position of a spot from spot data

    spot_data: numerical spot data
    actual: use the detected position if True, nominal position if False

    returns: a Point with x- and y-positions in pixels
    """
    prefix = "Pos" if actual else "Pos.Nom"
    x = int(spot_data[f"{prefix}.X"])
    y = int(spot_data[f"{prefix}.Y"])
    return Point(x, y)


def get_box(center, width, height=None):
    """returns the upper left and lower right corner of a bounding box

    center: position of the center of the box as "Point" namedtuple
    width: width of the bounding box
    height: height of the bounding box, if None, the width is used

    returns: tuple (upper left x, upper left y, lower right x, lower right y)
    """
    height = width if height is None else height

    dx = width / 2
    dy = height / 2

    # First, I tried to return (Point(x1, y1), Point(x2, y2)) but this format
    # is not accepted by all pillow methods that are used.
    # The form (x1, y1, x2, y2) is on the other hand.
    return (center.x - dx, center.y - dy, center.x + dx, center.y + dy)


def annotate_spot(canvas, spot_parameters, spot_data, scale=1):
    """annotate one spot position

    draws the circles and squares around a spot and adds the spot id

    canvas: a ImageDraw.Draw instance to use
    spot_parameters: settings of spot analysis
    spot_data: numerical spot data
    scale: image scale factor
    """

    found = spot_data["Spot.Found"]
    center = recalculate(get_position(spot_data, found), scale)
    color = GREEN if found else RED

    box = get_box(center, spot_parameters.roi_x, spot_parameters.roi_y)
    canvas.rectangle(box, outline=color, width=1)

    box = get_box(center, spot_parameters.radius_bkg)
    canvas.ellipse(box, outline=color, width=2)

    box = get_box(center, spot_parameters.radius_spot)
    canvas.ellipse(box, outline=color, width=1)

    canvas.text(center, str(spot_data["Pos.Id"]), RED, font=FONT)


def annotate_image(array_img, spot_parameters, array_data, scale):
    """annotates all spot positions in an image

    draws the circles and squares around all spots and adds the spot ids

    array_img: the image to annotate
    spot_parameters: settings of spot analysis
    array_data: numerical data of all spots
    scale: image scale factor
    """
    canvas = ImageDraw.Draw(array_img)
    for index, spot_data in array_data.iterrows():
        annotate_spot(canvas, spot_parameters, spot_data, scale)


def crop(array_img, spot_parameters, spot_data, scale=1):
    """creates a cropped image of one spot

    draws the circles and squares around all spots and adds the spot ids

    array_img: the image to crop
    spot_parameters: settings of spot analysis
    spot_data: numerical spot data
    scale: image scale factor

    returns: cropped image
    """
    found = spot_data["Spot.Found"]
    center = recalculate(get_position(spot_data, found), scale)
    box = get_box(center, spot_parameters.crop_x, spot_parameters.crop_y)
    return array_img.crop(box)
