from pathlib import Path
from collections import namedtuple

from defusedxml import ElementTree

ArrayParameters = namedtuple(
    "ArrayParameters", ["size_x", "size_y", "dist_x", "dist_y"]
)

SpotParameters = namedtuple(
    "SpotParameters",
    ["radius_spot", "radius_bkg", "roi_x", "roi_y", "crop_x", "crop_y"],
)


def _to_micro_meters(value):
    """helper method to convert a string with a value in mm to micrometers

    example: _to_micro_meters("0.350") == 350

    value: the value to convert

    returns: the value converted to micrometers
    """
    as_float = float(value)
    return int(as_float * 1000)


def array_parameters_from_xml(tree):
    """get the array parameter setting from an xml tree

    tree: xml tree

    returns: namedtuple ArrayParameters
    """
    layout = tree.find("Layout")
    sx = int(layout.attrib["NofSpotsX"])
    sy = int(layout.attrib["NofSpotsY"])
    dx = _to_micro_meters(layout.attrib["SpotDistMmX"])
    dy = _to_micro_meters(layout.attrib["SpotDistMmY"])
    return ArrayParameters(sx, sy, dx, dy)


def spot_parameters_from_xml(tree, array_parameters):
    """get the spot parameter setting from an xml tree

    Some spot parameters are set as array parameters.

    tree: xml tree
    array_parameters: namedtuple ArrayParameters

    returns: namedtuple SpotParameters
    """
    array = tree.find("MicroArrayAnalysis").find("Settings")
    rs = _to_micro_meters(array.attrib["MinSpotSizeMm"])
    rb = _to_micro_meters(array.attrib["MaxSpotSizeMm"])

    # this values are actually array parameters
    rx = array_parameters.dist_x
    ry = array_parameters.dist_y

    # before, the ROIMarginWidth and ROIMarginHeight settings from the array 
    # analysis was used. But this seems to be more or less independend of the
    # spot offset. Therfore we crop 100µm more than the spot offset. 
    # for future reference, how to access the node:
    # reg = tree.find("Registration").find("Settings")
    cx = rx + 100
    cy = ry + 100

    return SpotParameters(rs, rb, rx, ry, cx, cy)


def _parse_params_file(params_file, suffix):
    """parse a parameters setttings file

    For finding the params file in the raw data folder, the
    `sensospot_data._search_measurement_params_file()` method is used. The
    resulting file path is actually for a different file, but the one needed
    resides in the same folder

    params_file: one of the params file in the settings folder of the scan
    suffix: file name extension of the right xml file

    returns: ElementTree instance
    """
    xml_file = Path(params_file).with_suffix(suffix)
    with xml_file.open("r") as file_handle:
        tree = ElementTree.parse(file_handle)
    return tree


def get_array_parameters(params_file):
    """parse the array parameters file

    params_file: one of the params file in the settings folder of the scan

    returns: namedtuple ArrayParameters
    """
    tree = _parse_params_file(params_file, ".svary")
    return array_parameters_from_xml(tree)


def get_spot_parameters(params_file, array_parameters):
    """parse the array parameters file

    params_file: one of the params file in the settings folder of the scan
    array_parameters: namedtuple ArrayParameters

    returns: namedtuple SpotParameters
    """
    tree = _parse_params_file(params_file, ".svalg")
    return spot_parameters_from_xml(tree, array_parameters)
